import logging
import glob
import os
import sys
import gc

import pandas as pd
import anndata as ad
import numpy as np

from rich.progress import BarColumn, Progress
logger = logging.getLogger(__name__)


def create_fractions(no_celltypes):
    """
    Create random fractions
    :param no_celltypes: number of fractions to create
    :return: list of random fractions of length no_celltypes
    """
    fracs = np.random.rand(no_celltypes)
    fracs_sum = np.sum(fracs)
    fracs = np.divide(fracs, fracs_sum)
    return fracs


class BulkSimulator(object):
    """
    BulkSimulator class for the simulation of artificial bulk samples
    from scRNA-seq datasets

    :param sample_size: number of cells per sample
    :param num_samples: number of sample to simulate
    :param data_path: path to the data directory
    :param out_dir: output directory
    :param pattern of the data files
    :param unknown_celltypes: which celltypes to merge into the unknown class
    :param fmt: the format of the input files, can be txt or h5ad
    """

    def __init__(self, sample_size=100,
                 num_samples=1000,
                 data_path="./",
                 out_dir="./",
                 pattern="*_counts.txt",
                 unknown_celltypes=None,
                 fmt="txt"):
        if unknown_celltypes is None:
            unknown_celltypes = ["unknown"]

        self.sample_size = sample_size
        self.num_samples = num_samples // 2
        self.data_path = data_path
        self.out_dir = out_dir
        self.pattern = pattern
        self.unknown_celltypes = unknown_celltypes
        self.format = fmt
        self.datasets = []
        self.dataset_files = []

    def simulate(self):
        """ simulate artificial bulk datasets
        """

        # List available datasets
        if not self.data_path.endswith("/"):
            self.data_path += "/"
        files = glob.glob(os.path.join(self.data_path, self.pattern))
        files = [os.path.basename(x) for x in files]
        self.datasets = [x.replace(self.pattern.replace("*", ""), "") for x in files]
        self.dataset_files = [os.path.join(self.out_dir, x + ".h5ad") for x in self.datasets]

        if len(self.datasets) == 0:
            logging.error("No datasets found! Have you specified the pattern correctly?")
            sys.exit(1)

        logger.info("Datasets: [cyan]" + str(self.datasets) + "[/]")

        # Loop over datasets and simulate bulk data
        for i, dataset in enumerate(self.datasets):
            gc.collect()
            logger.info(f"[bold u]Simulating data from {dataset}")
            self.simulate_dataset(dataset)

        logger.info("[bold green]Finished data simulation!")

    def simulate_dataset(self, dataset):
        """
        Simulate bulk data from a single scRNA-seq dataset
        @param dataset:
        @type dataset:
        @return:
        @rtype:
        """

        # load the dataset
        data_x, data_y = self.load_dataset(dataset)

        # Merge unknown celltypes
        logger.info(f"Merging unknown cell types: {self.unknown_celltypes}")
        data_y = self.merge_unknown_celltypes(data_y)

        logger.info(f"Subsampling [bold cyan]{dataset}[/] ...")

        # Extract celltypes
        celltypes = list(set(data_y["Celltype"].tolist()))
        tmp_x, tmp_y = self.create_subsample_dataset(data_x, data_y, celltypes=celltypes)

        tmp_x = tmp_x.sort_index(axis=1)
        ratios = pd.DataFrame(tmp_y, columns=celltypes)
        ratios["ds"] = pd.Series(np.repeat(dataset, tmp_y.shape[0]), index=ratios.index)

        ann_data = ad.AnnData(X=tmp_x.to_numpy(), obs=ratios, var=pd.DataFrame(columns=[], index=list(tmp_x)))
        ann_data.uns['unknown'] = self.unknown_celltypes
        ann_data.uns['cell_types'] = celltypes

        ann_data.write(os.path.join(self.out_dir, dataset + ".h5ad"))

    def load_dataset(self, dataset):
        """
        Load a dataset
        @param dataset:
        @type dataset:
        @return:
        @rtype:
        """
        pattern = self.pattern.replace("*", "")
        logger.info(f"Loading [cyan]{dataset}[/] dataset ...")
        dataset_counts = dataset + pattern
        dataset_celltypes = dataset + "_celltypes.txt"

        # Load data in .txt format
        if self.format == 'txt':
            # Try to load celltypes
            try:
                y = pd.read_table(os.path.join(self.data_path, dataset_celltypes))
                # Check if has Celltype column
                if "Celltype" not in y.columns:
                    logger.error(
                        f"No 'Celltype' column found in {dataset}_celltypes.txt! Please make sure to include this "
                        f"column. "
                    )
                    sys.exit()
            except FileNotFoundError as e:
                logger.error(
                    f"No celltypes file found for [cyan]{dataset}[/]. It should be called [cyan]{dataset}_celltypes.txt."
                )
                sys.exit(e)

            # Try to load data file
            try:
                x = pd.read_table(os.path.join(self.data_path, dataset_counts), index_col=0)
            except FileNotFoundError as e:
                logger.error(
                    f"No counts file found for [cyan]{dataset}[/]. Was looking for file [cyan]{dataset_counts}[/]"
                )
                sys.exit(e)

            # Check that celltypes and count file have the same number of cells
            if not y.shape[0] == x.shape[0]:
                logger.error(
                    f"Different number of cells in {dataset}_celltypes and {dataset_counts}! Make sure the data has "
                    f"been processed correctly. "
                )
                sys.exit(1)

        # Load data in .h5ad format
        elif self.format == 'h5ad':
            try:
                data_h5ad = ad.read_h5ad(os.path.join(self.data_path, dataset_counts))
            except FileNotFoundError as e:
                logger.error(
                    f"No h5ad file found for [cyan]{dataset}[/]. Was looking for file [cyan]{dataset_counts}"
                )
                sys.exit(e)
            # cell types
            try:
                y = pd.DataFrame(data_h5ad.obs.Celltype)
                y.reset_index(inplace=True, drop=True)
            except Exception as e:
                logger.error(
                    f"Celltype attribute not found for [cyan]{dataset}"
                )
                sys.exit(e)
            # counts
            x = pd.DataFrame(data_h5ad.X.todense())
            x.index = data_h5ad.obs_names
            x.columns = data_h5ad.var_names
            del data_h5ad
        else:
            logger.error(
                f"Unsupported file format {self.format}!"
            )
            sys.exit(1)

        return x, y

    def merge_unknown_celltypes(self, y):
        """
        Merge all unknown celltypes together
        :param y: list of cell type labels
        :return: list of cell types with unknown cell types merged into "Unknown" type
        """
        celltypes = list(y["Celltype"])
        new_celltypes = ["Unknown" if x in self.unknown_celltypes else x for x in celltypes]
        y["Celltype"] = new_celltypes
        return y

    def create_subsample_dataset(self, x, y, celltypes):
        """
        Generate many artifial bulk samples with known fractions
        This function will create normal and sparse samples (no_samples)
        @param x:
        @param y:
        @param celltypes:
        @return:
        """
        sim_x = []
        sim_y = []

        # Create normal samples
        progress_bar = Progress(
            "[bold blue]{task.description}",
            "[bold cyan]{task.fields[samples]}",
            BarColumn(bar_width=None)
        )
        with progress_bar:
            normal_samples_progress = progress_bar.add_task("Normal samples", total=self.num_samples, samples=0)
            sparse_samples_progress = progress_bar.add_task("Sparse samples", total=self.num_samples, samples=0)

            # Create normal samples
            for i in range(self.num_samples):
                progress_bar.update(normal_samples_progress, advance=1, samples=i+1)
                sample, label = self.create_subsample(x, y, celltypes)
                sim_x.append(sample)
                sim_y.append(label)

            # Create sparase samples
            for i in range(self.num_samples):
                progress_bar.update(sparse_samples_progress, advance=1, samples=i+1)
                sample, label = self.create_subsample(x, y, celltypes, sparse=True)
                sim_x.append(sample)
                sim_y.append(label)

        sim_x = pd.concat(sim_x, axis=1).T
        sim_y = pd.DataFrame(sim_y, columns=celltypes)

        return sim_x, sim_y

    def create_subsample(self, x, y, celltypes, sparse=False):
        """
        Generate artifical bulk subsample with random fractions of celltypes
        If sparse is set to true, add random celltypes to the missing celltypes

        @param x:
        @param y:
        @param celltypes:
        @param sparse:
        @return:
        """
        available_celltypes = celltypes
        if sparse:
            no_keep = np.random.randint(1, len(available_celltypes))
            keep = np.random.choice(
                list(range(len(available_celltypes))), size=no_keep, replace=False
            )
            available_celltypes = [available_celltypes[i] for i in keep]

        no_avail_cts = len(available_celltypes)

        # Create fractions for available celltypes
        fracs = create_fractions(no_celltypes=no_avail_cts)
        samp_fracs = np.multiply(fracs, self.sample_size)
        samp_fracs = list(map(int, samp_fracs))

        # Make complete fracions
        fracs_complete = [0] * len(celltypes)
        for i, act in enumerate(available_celltypes):
            idx = celltypes.index(act)
            fracs_complete[idx] = fracs[i]

        artificial_samples = []
        for i in range(no_avail_cts):
            ct = available_celltypes[i]
            cells_sub = x.loc[np.array(y["Celltype"] == ct), :]
            cells_fraction = np.random.randint(0, cells_sub.shape[0], samp_fracs[i])
            cells_sub = cells_sub.iloc[cells_fraction, :]
            artificial_samples.append(cells_sub)

        df_samp = pd.concat(artificial_samples, axis=0)
        df_samp = df_samp.sum(axis=0)

        return df_samp, fracs_complete

    @staticmethod
    def merge_datasets(data_dir="./", files=None, out_name="data.h5ad"):
        """

        @param out_name: name of the merged .h5ad file
        @param data_dir: directory to look for datasets
        @param files: list of files to merge
        @return:
        """
        non_celltype_obs = ["ds", "batch"]
        if not files:
            files = glob.glob(os.path.join(data_dir, "*.h5ad"))

        logger.info(f"Merging datasets: {files} into [bold cyan]{out_name}")

        # load first file
        adata = ad.read_h5ad(files[0])

        for i in range(1, len(files)):
            adata = adata.concatenate(ad.read_h5ad(files[i]), uns_merge="same")

        combined_celltypes = list(adata.obs.columns)
        combined_celltypes = [x for x in combined_celltypes if not x in non_celltype_obs]
        for ct in combined_celltypes:
            adata.obs[ct].fillna(0, inplace=True)

        adata.uns['cell_types'] = combined_celltypes
        adata.write(out_name)



"""
scaden Main functionality

Contains code to
- process a training datasets
- train a model
- perform predictions

"""

# Imports
import logging
from scaden.model.architectures import architectures
from scaden.model.scaden import Scaden


logger = logging.getLogger(__name__)

"""
PARAMETERS
"""
# ==========================================#

# Extract architectures
M256_HIDDEN_UNITS = architectures["m256"][0]
M512_HIDDEN_UNITS = architectures["m512"][0]
M1024_HIDDEN_UNITS = architectures["m1024"][0]
M256_DO_RATES = architectures["m256"][1]
M512_DO_RATES = architectures["m512"][1]
M1024_DO_RATES = architectures["m1024"][1]

# ==========================================#


def training(
    data_path, train_datasets, model_dir, batch_size, learning_rate, num_steps, seed=0
):
    """
    Perform training of three a scaden model ensemble consisting of three different models
    :param model_dir:
    :param batch_size:
    :param learning_rate:
    :param num_steps:
    :return:
    """
    # Convert training datasets
    if train_datasets == "":
        train_datasets = []
    else:
        train_datasets = train_datasets.split(",")
        logger.info(f"Training on: [cyan]{train_datasets}")

    # Training of M256 model
    logger.info("[cyan]Training M256 Model ... [/]")
    cdn256 = Scaden(
        model_dir=model_dir + "/m256",
        model_name="m256",
        batch_size=batch_size,
        learning_rate=learning_rate,
        num_steps=num_steps,
        seed=seed,
        hidden_units=M256_HIDDEN_UNITS,
        do_rates=M512_DO_RATES,
    )
    cdn256.train(input_path=data_path, train_datasets=train_datasets)
    del cdn256

    # Training of M512 model
    logger.info("[cyan]Training M512 Model ... [/]")
    cdn512 = Scaden(
        model_dir=model_dir + "/m512",
        model_name="m512",
        batch_size=batch_size,
        learning_rate=learning_rate,
        num_steps=num_steps,
        seed=seed,
        hidden_units=M512_HIDDEN_UNITS,
        do_rates=M512_DO_RATES,
    )
    cdn512.train(input_path=data_path, train_datasets=train_datasets)
    del cdn512

    # Training of M1024 model
    logger.info("[cyan]Training M1024 Model ... [/]")
    cdn1024 = Scaden(
        model_dir=model_dir + "/m1024",
        model_name="m1024",
        batch_size=batch_size,
        learning_rate=learning_rate,
        num_steps=num_steps,
        seed=seed,
        hidden_units=M1024_HIDDEN_UNITS,
        do_rates=M1024_DO_RATES,
    )
    cdn1024.train(input_path=data_path, train_datasets=train_datasets)
    del cdn1024

    logger.info("[green]Training finished.")


"""
Generate random example data which allows for testing and
to give users examples for the input format
"""
import random
import os
import logging
import pandas as pd
import numpy as np
import sys

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def exampleData(n_cells=10, n_genes=100, n_samples=10, n_types=5, out_dir="./"):
    """
    Generate an example scRNA-seq count file
    :param n: number of cells
    :param g: number of genes
    """
    if n_types > n_cells:
        logger.error("You can't specifiy more cell types than cells!")
        sys.exit(1)

    # Generate example scRNA-seq data
    counts = np.random.randint(low=0, high=1000, size=(n_cells, n_genes))
    gene_names = ["gene"] * n_genes
    for i in range(len(gene_names)):
        gene_names[i] = gene_names[i] + str(i)
    df = pd.DataFrame(counts, columns=gene_names)

    # Generate example celltype labels
    celltypes = ["celltype"] * np.random.randint(n_types)
    for i in range(len(celltypes)):
        celltypes[i] = celltypes[i] + str(i)
    celltype_list = random.choices(celltypes, k=n_cells)
    ct_df = pd.DataFrame(celltype_list, columns=["Celltype"])

    # Generate example bulk RNA-seq data
    bulk = np.random.randint(low=0, high=1000, size=(n_genes, n_samples))
    samples = ["sample"] * n_samples
    for i in range(len(samples)):
        samples[i] = samples[i] + str(i)
    bulk_df = pd.DataFrame(bulk, columns=samples, index=gene_names)

    # Save the data
    df.to_csv(os.path.join(out_dir, "example_counts.txt"), sep="\t")
    ct_df.to_csv(os.path.join(out_dir, "example_celltypes.txt"), sep="\t", index=False)
    bulk_df.to_csv(os.path.join(out_dir, "example_bulk_data.txt"), sep="\t")

    logger.info(f"Example data has been created in [cyan]{out_dir}")



import click
import sys
import scaden
import rich
import rich.logging
import logging
import os
import tensorflow as tf
from scaden.train import training
from scaden.predict import prediction
from scaden.process import processing
from scaden.simulate import simulation
from scaden.example import exampleData
from scaden.merge import merge_datasets
"""

author: Kevin Menden

This is the main file for executing the Scaden program.
"""

# Logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)
logger.addHandler(
    rich.logging.RichHandler(
        level=logging.INFO,
        console=rich.console.Console(file=sys.stderr),
        show_time=False,
        markup=True,
    )
)

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"


def main():
    text = """
     ____                _            
    / ___|  ___ __ _  __| | ___ _ __  
    \___ \ / __/ _` |/ _` |/ _ \ '_ \ 
     ___) | (_| (_| | (_| |  __/ | | |
    |____/ \___\__,_|\__,_|\___|_| |_|

    """
    click.echo(click.style(text, fg="blue"))
    cli()


if __name__ == "__main__":
    main()
"""
Set up the command line client with different commands to execute
"""


@click.group()
@click.version_option(scaden.__version__)
def cli():
    pass


"""
Training mode
"""


@cli.command()
@click.argument(
    "data_path", type=click.Path(exists=True), required=True, metavar="<training data>"
)
@click.option(
    "--train_datasets",
    default="",
    help="Comma-separated list of datasets used for training. Uses all by default.",
)
@click.option("--model_dir", default="./", help="Path to store the model in")
@click.option(
    "--batch_size", default=128, help="Batch size to use for training. Default: 128"
)
@click.option(
    "--learning_rate",
    default=0.0001,
    help="Learning rate used for training. Default: 0.0001",
)
@click.option("--steps", default=5000, help="Number of training steps")
@click.option("--seed", default=0, help="Set random seed")
def train(data_path, train_datasets, model_dir, batch_size, learning_rate, steps, seed):
    """ Train a Scaden model """
    training(
        data_path=data_path,
        train_datasets=train_datasets,
        model_dir=model_dir,
        batch_size=batch_size,
        learning_rate=learning_rate,
        num_steps=steps,
        seed=seed,
    )


"""
Prediction mode
"""


@cli.command()
@click.argument(
    "data_path",
    type=click.Path(exists=True),
    required=True,
    metavar="<prediction data>",
)
@click.option("--model_dir", default="./", help="Path to trained model")
@click.option(
    "--outname", default="scaden_predictions.txt", help="Name of predictions file."
)
@click.option("--seed", default=0, help="Set random seed")
def predict(data_path, model_dir, outname, seed):
    """ Predict cell type composition using a trained Scaden model"""
    prediction(model_dir=model_dir, data_path=data_path, out_name=outname, seed=seed)


"""
Processing mode
"""


@cli.command()
@click.argument(
    "data_path",
    type=click.Path(exists=True),
    required=True,
    metavar="<training dataset to be processed>",
)
@click.argument(
    "prediction_data",
    type=click.Path(exists=True),
    required=True,
    metavar="<data for prediction>",
)
@click.option(
    "--processed_path",
    default="processed.h5ad",
    help="Path of processed file. Must end with .h5ad",
)
@click.option(
    "--var_cutoff",
    default=0.1,
    help="Filter out genes with a variance less than the specified cutoff. A low cutoff is recommended,"
         "this should only remove genes that are obviously uninformative.",
)
def process(data_path, prediction_data, processed_path, var_cutoff):
    """ Process a dataset for training """
    processing(
        data_path=prediction_data,
        training_data=data_path,
        processed_path=processed_path,
        var_cutoff=var_cutoff,
    )


"""
Simulate dataset
"""


@cli.command()
@click.option("--out", "-o", default="./", help="Directory to store output files in")
@click.option("--data", "-d", default=".", help="Path to scRNA-seq dataset(s)")
@click.option(
    "--cells", "-c", default=100, help="Number of cells per sample [default: 100]"
)
@click.option(
    "--n_samples",
    "-n",
    default=1000,
    help="Number of samples to simulate [default: 1000]",
)
@click.option(
    "--pattern",
    default="*_counts.txt",
    help="File pattern to recognize your processed scRNA-seq count files",
)
@click.option(
    "--unknown",
    "-u",
    multiple=True,
    default=["unknown"],
    help="Specifiy cell types to merge into the unknown category. Specify this flag for every cell type you want to "
         "merge in unknown. [default: unknown]",
)
@click.option(
    "--prefix",
    "-p",
    default="data",
    help="Prefix to append to training .h5ad file [default: data]",
)
@click.option(
    "--data-format",
    "-f",
    default="txt",
    help="Data format of scRNA-seq data, can be 'txt' or 'h5ad' [default: 'txt']",
)
def simulate(out, data, cells, n_samples, pattern, unknown, prefix, data_format):
    """ Create artificial bulk RNA-seq data from scRNA-seq dataset(s)"""
    simulation(
        simulate_dir=out,
        data_dir=data,
        sample_size=cells,
        num_samples=n_samples,
        pattern=pattern,
        unknown_celltypes=unknown,
        out_prefix=prefix,
        fmt=data_format
    )


"""
Merge simulated datasets
"""


@cli.command()
@click.option("--data", "-d", default=".", help="Directory containing simulated datasets (in .h5ad format)")
@click.option("--prefix", "-p", default="data", help="Prefix of output file [default: data]")
@click.option("--files", "-f", default=None, help="Comma-separated list of filenames to merge")
def merge(data, prefix, files):
    """ Merge simulated datasets into on training dataset """
    merge_datasets(data_dir=data, prefix=prefix, files=files)


"""
Generate example data
"""


@cli.command()
@click.option("--cells", "-c", default=10, help="Number of cells [default: 10]")
@click.option("--types", "-t", default=5, help="Number of cell types [default: 5]")
@click.option("--genes", "-g", default=100, help="Number of genes [default: 100]")
@click.option("--out", "-o", default="./", help="Output directory [default: ./]")
@click.option(
    "--samples", "-n", default=10, help="Number of bulk samples [default: 10]"
)
def example(cells, genes, samples, out, types):
    """ Generate an example dataset """
    exampleData(n_cells=cells, n_genes=genes, n_samples=samples, out_dir=out, n_types=types)



"""
scaden Main functionality

Contains code to
- process a training datasets
- train a model
- perform predictions

"""

# Imports
import logging
import tensorflow as tf
from anndata import read_h5ad
from scaden.model.architectures import architectures
from scaden.model.scaden import Scaden

logger = logging.getLogger(__name__)

"""
PARAMETERS
"""
# ==========================================#

# Extract architectures
M256_HIDDEN_UNITS = architectures["m256"][0]
M512_HIDDEN_UNITS = architectures["m512"][0]
M1024_HIDDEN_UNITS = architectures["m1024"][0]
M256_DO_RATES = architectures["m256"][1]
M512_DO_RATES = architectures["m512"][1]
M1024_DO_RATES = architectures["m1024"][1]

# ==========================================#


def prediction(model_dir, data_path, out_name, seed=0):
    """
    Perform prediction using a trained scaden ensemble
    :param model_dir: the directory containing the models
    :param data_path: the path to the gene expression file
    :param out_name: name of the output prediction file
    :return:
    """

    # Small model predictions
    cdn256 = Scaden(
        model_dir=model_dir + "/m256",
        model_name="m256",
        seed=seed,
        hidden_units=M256_HIDDEN_UNITS,
        do_rates=M256_DO_RATES,
    )
    # Predict ratios
    preds_256 = cdn256.predict(
        input_path=data_path
    )

    # Mid model predictions
    cdn512 = Scaden(
        model_dir=model_dir + "/m512",
        model_name="m512",
        seed=seed,
        hidden_units=M512_HIDDEN_UNITS,
        do_rates=M512_DO_RATES,
    )
    # Predict ratios
    preds_512 = cdn512.predict(
        input_path=data_path
    )

    # Large model predictions
    cdn1024 = Scaden(
        model_dir=model_dir + "/m1024",
        model_name="m1024",
        seed=seed,
        hidden_units=M1024_HIDDEN_UNITS,
        do_rates=M256_DO_RATES,
    )
    # Predict ratios
    preds_1024 = cdn1024.predict(
        input_path=data_path
    )

    # Average predictions
    preds = (preds_256 + preds_512 + preds_1024) / 3
    preds.to_csv(out_name, sep="\t")
    logger.info(f"[bold]Created cell composition predictions: [green]{out_name}[/]")



"""
scaden Main functionality

Contains code to
- process a training datasets
- train a model
- perform predictions

"""

# Imports
import tensorflow as tf
from anndata import read_h5ad
from scaden.model.functions import get_signature_genes, preprocess_h5ad_data

"""
PARAMETERS
"""
# ==========================================#

def processing(data_path, training_data, processed_path, var_cutoff):
    """
    Process a training dataset to contain only the genes also available in the prediction data
    :param data_path: path to prediction data
    :param training_data: path to training data (h5ad file)
    :param processed_path: name of processed file
    :return:
    """
    # Get the common genes (signature genes)
    raw_input = read_h5ad(training_data)
    sig_genes_complete = list(raw_input.var_names)
    sig_genes = get_signature_genes(input_path=data_path, sig_genes_complete=sig_genes_complete, var_cutoff=var_cutoff)

    # Pre-process data with new signature genes
    preprocess_h5ad_data(raw_input_path=training_data,
                         processed_path=processed_path,
                         sig_genes=sig_genes)



#!/usr/bin/env python3

from setuptools import setup, find_packages

version = "1.1.0"


with open("README.md", "r", encoding="UTF-8") as fh:
    long_description = fh.read()

with open("LICENSE", encoding="UTF-8") as f:
    license = f.read()

setup(
    name="scaden",
    version=version,
    description="Cell type deconvolution using single cell data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=[
        "bioinformatics",
        "deep learning",
        "machine learning",
        "single cell sequencing",
        "deconvolution",
    ],
    author="Kevin Menden",
    author_email="kevin.menden@t-online.de",
    url="https://github.com/KevinMenden/scaden",
    license="MIT License",
    entry_points={"console_scripts": ["scaden=scaden.__main__:main"]},
    packages=find_packages(),
    include_package_data=True,
    python_requires=">3.6.0",
    install_requires=[
        "pandas",
        "numpy",
        "scikit-learn",
        "tensorflow>=2.0",
        "anndata",
        "click",
        "h5py~=2.10.0",
        "rich",
    ],
)



from scaden.simulation import BulkSimulator

"""
Merge simulate datasets
"""


def merge_datasets(data_dir, prefix, files=None):

    bulk_simulator = BulkSimulator()

    if files:
        files = files.split(",")

    # Merge the resulting datasets
    bulk_simulator.merge_datasets(data_dir=data_dir,
                                  files=files,
                                  out_name=prefix + ".h5ad")



from scaden.simulation import BulkSimulator

"""
Simulation of artificial bulk RNA-seq samples from scRNA-seq data
and subsequenbt formatting in .h5ad file for training with Scaden
"""


def simulation(simulate_dir, data_dir, sample_size, num_samples, pattern,
               unknown_celltypes, out_prefix, fmt):

    unknown_celltypes = list(unknown_celltypes)
    bulk_simulator = BulkSimulator(sample_size=sample_size,
                                   num_samples=num_samples,
                                   data_path=data_dir,
                                   out_dir=simulate_dir,
                                   pattern=pattern,
                                   unknown_celltypes=unknown_celltypes,
                                   fmt=fmt)

    # Perform dataset simulation
    bulk_simulator.simulate()

    # Merge the resulting datasets
    bulk_simulator.merge_datasets(data_dir=simulate_dir,
                                  files=bulk_simulator.dataset_files,
                                  out_name=out_prefix + ".h5ad")



"""
Cell Deconvolutional Network (scaden) class
"""
import os
import logging
import sys
import gc
import tensorflow as tf
import numpy as np
import pandas as pd
from anndata import read_h5ad
import collections
from .functions import sample_scaling
from rich.progress import Progress, BarColumn

logger = logging.getLogger(__name__)
tf.get_logger().setLevel('ERROR')
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

class Scaden(object):
    """
    scaden class
    """

    def __init__(
        self,
        model_dir,
        model_name,
        batch_size=128,
        learning_rate=0.0001,
        num_steps=1000,
        seed=0,
        hidden_units=[256, 128, 64, 32],
        do_rates=[0, 0, 0, 0],
    ):

        self.model_dir = model_dir
        self.batch_size = batch_size
        self.model_name = model_name
        self.beta1 = 0.9
        self.beta2 = 0.999
        self.learning_rate = learning_rate
        self.data = None
        self.n_classes = None
        self.labels = None
        self.x = None
        self.y = None
        self.num_steps = num_steps
        self.scaling = "log_min_max"
        self.sig_genes = None
        self.sample_names = None
        self.hidden_units = hidden_units
        self.do_rates = do_rates

        # Set seeds for reproducibility
        tf.random.set_seed(seed)
        os.environ["TF_DETERMINISTIC_OPS"] = "1"
        np.random.seed(seed)

    def scaden_model(self, n_classes):
        """Create the Scaden model"""

        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Dense(self.hidden_units[0], activation=tf.nn.relu))
        model.add(tf.keras.layers.Dropout(self.do_rates[0]))
        model.add(tf.keras.layers.Dense(self.hidden_units[1], activation=tf.nn.relu))
        model.add(tf.keras.layers.Dropout(self.do_rates[1]))
        model.add(tf.keras.layers.Dense(self.hidden_units[2], activation=tf.nn.relu))
        model.add(tf.keras.layers.Dropout(self.do_rates[2]))
        model.add(tf.keras.layers.Dense(self.hidden_units[3], activation=tf.nn.relu))
        model.add(tf.keras.layers.Dropout(self.do_rates[3]))
        model.add(tf.keras.layers.Dense(n_classes, activation=tf.nn.softmax))

        return model

    def compute_loss(self, logits, targets):
        """
        Compute L1 loss
        :param logits:
        :param targets:
        :return: L1 loss
        """
        loss = tf.reduce_mean(input_tensor=tf.math.square(logits - targets))
        return loss

    def compute_accuracy(self, logits, targets, pct_cut=0.05):
        """
        Compute prediction accuracy
        :param targets:
        :param pct_cut:
        :return:
        """
        equality = tf.less_equal(
            tf.math.abs(tf.math.subtract(logits, targets)), pct_cut
        )
        accuracy = tf.reduce_mean(input_tensor=tf.cast(equality, tf.float32))
        return accuracy

    def correlation_coefficient(self, logits, targets):
        """
        Calculate the pearson correlation coefficient
        :param logits:
        :param targets:
        :return:
        """
        mx = tf.reduce_mean(input_tensor=logits)
        my = tf.reduce_mean(input_tensor=targets)
        xm, ym = logits - mx, targets - my
        r_num = tf.reduce_sum(input_tensor=tf.multiply(xm, ym))
        r_den = tf.sqrt(
            tf.multiply(
                tf.reduce_sum(input_tensor=tf.square(xm)),
                tf.reduce_sum(input_tensor=tf.square(ym)),
            )
        )
        r = tf.divide(r_num, r_den)
        r = tf.maximum(tf.minimum(r, 1.0), -1.0)
        return r

    def visualization(self, logits, targets, classes):
        """
        Create evaluation metrics
        :param targets:
        :param classes:
        :return:
        """
        # add evaluation metrics
        rmse = tf.compat.v1.metrics.root_mean_squared_error(logits, targets)[1]
        pcor = self.correlation_coefficient(logits, targets)
        eval_metrics = {"rmse": rmse, "pcor": pcor}

        for i in range(logits.shape[1]):
            eval_metrics[
                "mre_" + str(classes[i])
            ] = tf.compat.v1.metrics.mean_relative_error(
                targets[:, i], logits[:, i], targets[:, i]
            )[
                0
            ]
            eval_metrics[
                "mae_" + str(classes[i])
            ] = tf.compat.v1.metrics.mean_absolute_error(
                targets[:, i], logits[:, i], targets[:, i]
            )[
                0
            ]
            eval_metrics["pcor_" + str(classes[i])] = self.correlation_coefficient(
                targets[:, i], logits[:, i]
            )

        eval_metrics["mre_total"] = tf.compat.v1.metrics.mean_relative_error(
            targets, logits, targets
        )[1]

        eval_metrics["mae_total"] = tf.compat.v1.metrics.mean_relative_error(
            targets, logits, targets
        )[1]

        eval_metrics["accuracy01"] = self.compute_accuracy(
            logits, targets, pct_cut=0.01
        )
        eval_metrics["accuracy05"] = self.compute_accuracy(
            logits, targets, pct_cut=0.05
        )
        eval_metrics["accuracy1"] = self.compute_accuracy(logits, targets, pct_cut=0.1)

        # Create summary scalars
        for key, value in eval_metrics.items():
            tf.compat.v1.summary.scalar(key, value)

        tf.compat.v1.summary.scalar("loss", self.loss)

        merged_summary_op = tf.compat.v1.summary.merge_all()

        return merged_summary_op

    def load_h5ad_file(self, input_path, batch_size, datasets=[]):
        """
        Load input data from a h5ad file and divide into training and test set
        :param input_path: path to h5ad file
        :param batch_size: batch size to use for training
        :param datasets: a list of datasets to extract from the file
        :return: Dataset object
        """
        try:
            raw_input = read_h5ad(input_path)
        except:
            logger.error(
                "Could not load training data file! Is it a .h5ad file generated with `scaden process`?"
            )
            sys.exit()

        # Subset dataset if --train_datasets is given
        if len(datasets) > 0:
            all_ds = collections.Counter(raw_input.obs["ds"])

            # Check that given datasets are all actually available
            for ds in datasets:
                if not ds in all_ds:
                    logger.warn(
                        f"The dataset '[cyan]{ds}[/cyan]' could not be found in the training data! Is the name correct?"
                    )

            for ds in all_ds:
                if ds not in datasets:
                    raw_input = raw_input[raw_input.obs["ds"] != ds].copy()

        # Create training dataset
        ratios = [raw_input.obs[ctype] for ctype in raw_input.uns["cell_types"]]
        self.x_data = raw_input.X.astype(np.float32)
        self.y_data = np.array(ratios, dtype=np.float32).transpose()
        self.data = tf.data.Dataset.from_tensor_slices((self.x_data, self.y_data))
        self.data = self.data.shuffle(1000).repeat().batch(batch_size=batch_size)
        self.data_iter = iter(self.data)

        # Extract celltype and feature info
        self.labels = raw_input.uns["cell_types"]
        self.sig_genes = list(raw_input.var_names)

    def load_prediction_file(self, input_path, sig_genes, labels, scaling=None):
        """
        Load a file to perform prediction on it
        :param input_path: path to input file
        :param sig_genes: the signature genes to use
        :param scaling: which scaling to perform
        :return: Dataset object
        """
        # Load data
        data = pd.read_table(input_path, sep="\t", index_col=0)
        sample_names = list(data.columns)

        # check for duplicates
        data_index = list(data.index)
        if not (len(data_index) == len(set(data_index))):
            logger.warn(
                "Scaden Warning: Your mixture file conatins duplicate genes! The first occuring gene will be used for every duplicate."
            )
            data = data.loc[~data.index.duplicated(keep="first")]

        data = data.loc[sig_genes]
        data = data.T

        # Scaling
        if scaling:
            data = sample_scaling(data, scaling_option=scaling)

        self.data = data

        return sample_names

    def build_model(self, input_path, train_datasets, mode="train"):
        """
        Build the model graph
        :param reuse:
        :return:
        """
        self.global_step = tf.Variable(0, name="global_step", trainable=False)

        # Load training data
        if mode == "train":
            self.load_h5ad_file(
                input_path=input_path,
                batch_size=self.batch_size,
                datasets=train_datasets,
            )

        # Load prediction data
        if mode == "predict":
            self.sample_names = self.load_prediction_file(
                input_path=input_path,
                sig_genes=self.sig_genes,
                labels=self.labels,
                scaling=self.scaling,
            )

        # Build the model or load if available
        self.n_classes = len(self.labels)

        try:
            self.model = tf.keras.models.load_model(self.model_dir, compile=False)
            logger.info(f"Loaded pre-trained model: [cyan]{self.model_name}")
        except:
            self.model = self.scaden_model(n_classes=self.n_classes)

    def train(self, input_path, train_datasets):
        """
        Train the model
        :param num_steps:
        :return:
        """

        # Define the optimizer
        optimizer = tf.keras.optimizers.Adam(learning_rate=self.learning_rate)

        # Build model graph
        self.build_model(
            input_path=input_path, train_datasets=train_datasets, mode="train"
        )

        # Training loop
        progress_bar = Progress(
            "[bold blue]{task.description}",
            "[bold cyan]Step: {task.fields[step]}, Loss: {task.fields[loss]}",
            BarColumn(bar_width=None),
        )

        training_progress = progress_bar.add_task(self.model_name, total=self.num_steps, step=0, loss=1)
        with progress_bar:

            for step in range(self.num_steps):

                x, y = self.data_iter.get_next()

                with tf.GradientTape() as tape:
                    self.logits = self.model(x, training=True)
                    loss = self.compute_loss(self.logits, y)

                grads = tape.gradient(loss, self.model.trainable_weights)

                optimizer.apply_gradients(zip(grads, self.model.trainable_weights))

                progress_bar.update(training_progress, advance=1, step=step, loss=f"{loss:.4f}")

                # Collect garbage after 100 steps - otherwise runs out of memory
                if step % 100 == 0:
                    gc.collect()


        # Save the trained model
        self.model.save(self.model_dir)
        pd.DataFrame(self.labels).to_csv(
            os.path.join(self.model_dir, "celltypes.txt"), sep="\t"
        )
        pd.DataFrame(self.sig_genes).to_csv(
            os.path.join(self.model_dir, "genes.txt"), sep="\t"
        )

    def predict(self, input_path):
        """
        Perform prediction with a pre-trained model
        :param input_path: prediction data path
        :return:
        """
        # Load signature genes and celltype labels
        sig_genes = pd.read_table(self.model_dir + "/genes.txt", index_col=0)
        self.sig_genes = list(sig_genes["0"])
        labels = pd.read_table(self.model_dir + "/celltypes.txt", index_col=0)
        self.labels = list(labels["0"])

        # Build model graph
        self.build_model(input_path=input_path, train_datasets=[], mode="predict")

        predictions = self.model.predict(self.data)

        pred_df = pd.DataFrame(
            predictions, columns=self.labels, index=self.sample_names
        )
        return pred_df
